import debugpy
import logging
import threading
from .threads import StoppableThread


class DebugPyEx():
    """Use this class instead of debugpy.
    It provides an additional manageability layer on top of debugpy calls.
    """
    _debugpy_connected = False
    _connect_lock = threading.Lock()

    def _thread_connect_proc(host, port):
        try:
            debugpy.connect((str(host), int(port)))
            DebugPyEx._debugpy_connected = True
        except SystemExit:
            # SystemExit is a "legal" way to terminate this thread.
            logging.warn("Debugpy thread has been terminated.")


    @staticmethod
    def connect(host, port, connect_timeout_seconds) -> bool:
        with DebugPyEx._connect_lock:
            DebugPyEx._debugpy_connected = False
            thread = StoppableThread(target=DebugPyEx._thread_connect_proc, args=(
                host, port,), daemon=True)
            thread.start()
            thread.join(connect_timeout_seconds)
            if(thread.is_alive()):
                # kill the thread "gracefully"!
                thread.stop()
                return False
            elif DebugPyEx._debugpy_connected:
                debugpy.debug_this_thread()
                return True
            else:
                return False




from enum import Enum
import os
import signal
import argparse
import sys
import logging
import subprocess
import stat
import threading
import urllib.request
from pathlib import Path
import ssl
import platform
import tarfile
import time
import typing
import json
import zipfile

class DebugMode(Enum):
    """Debugging mode enum:
    waiting for another machine to connect, or connect to another machine
    """
    # Start a remote forwarder with Azure Relay Bridge, another side is attaching
    WaitForConnection=1,
    Listen = WaitForConnection,
    # Start a local forwarder with Azure Relay Bridge, another side is listening
    Connect = 2,
    Attach = Connect


class DebugRelay(object):
    """Initializes and controls Azure Relay Bridge process.

    Raises:
        ValueError: Invalid arguments.
        TimeoutError: Azure Relay Bridge took too long to connect.
    """
    # Azure Relay Bridge executable name
    relay_app_name = "azbridge"
    # `~/.azdebugrelay` installation directory
    relay_dir_name = ".azdebugrelay"
    # current Azure Debug Relay build
    relay_version_name = "0.2.9"
    # are we running on Windows?
    is_windows = platform.platform().lower().startswith("windows")

    DEFAULT_AZ_RELAY_BRIDGE_UBUNTU_DOWLOAD =\
        "https://github.com/vladkol/azure-relay-bridge/releases/download/v0.2.9/azbridge.azrelay_folder-rel.ubuntu.18.04-x64.tar.gz"
    DEFAULT_AZ_RELAY_BRIDGE_MACOS_DOWLOAD =\
        "https://github.com/vladkol/azure-relay-bridge/releases/download/v0.2.9/azbridge.0.2.9-rel.osx-x64.tar.gz"
    DEFAULT_AZ_RELAY_BRIDGE_DEBIAN_DOWLOAD =\
        "https://github.com/vladkol/azure-relay-bridge/releases/download/v0.2.9/azbridge.0.2.9-rel.debian.10-x64.tar.gz"
    DEFAULT_AZ_RELAY_BRIDGE_WINDOWS_DOWLOAD =\
        "https://github.com/vladkol/azure-relay-bridge/releases/download/v0.2.9/azbridge.0.2.9-rel.win10-x64.zip"

    _installed_az_relay = False
    _relay_config_file = None


    def __init__(self,
                 access_key_or_connection_string: str,
                 relay_connection_name: str,
                 debug_mode: DebugMode = DebugMode.WaitForConnection,
                 hybrid_connection_url: str = None,
                 host: str ="127.0.0.1",
                 ports: typing.List[str] = ["5678"],
                 az_relay_connection_wait_time: float = 60,
                 logger: logging.Logger = logging.root):
        """Initializes DebugRelay object. 
        
        Args:
            access_key_or_connection_string (str): access key or connection string for Azure Relay Hybrid Connection
            relay_connection_name (str): name of Azure Relay Hybrid Connection
            debug_mode (DebugMode, optional): Connect or Listen (WaitForConnection). Defaults to DebugMode.WaitForConnection.
            hybrid_connection_url (str, optional): optional URL of Hybrid Connection. Defaults to None. 
                Required when access_key_or_connection_string is an access key.
            host (str, optional): Local hostname/address the debugging starts on. Defaults to "127.0.0.1".
            ports (typing.List[str], optional): Any available ports that you can use within your machine.
                This port will be connected to or exposed by Azure Relay Bridge. Defaults to ["5678"].
            az_relay_connection_wait_time (float, optional): Maximum time to wait for Azure Relay Bridge
                to initialize and connect when open() is called with wait_for_connection == True. Defaults to 60.

        Raises:
            ValueError: hybrid_connection_url is None while access_key_or_connection_string is not a connection string.
        """
        self.logger = logger

        self.relay_subprocess = None
        if access_key_or_connection_string.startswith("Endpoint="):
            have_connection_string = True
        else:
            have_connection_string = False
        if hybrid_connection_url is None or hybrid_connection_url == "":
            if not have_connection_string:
                raise ValueError(
                    "hybrid_connection_url must be specified when "\
                    "access_key_or_connection_string is not a connection string.")

        if have_connection_string:
            self.auth_option = f"-x \"{access_key_or_connection_string}\"" 
        else:
            self.auth_option = f"-E \"{hybrid_connection_url}\" -k \"{access_key_or_connection_string}\""

        if debug_mode == DebugMode.WaitForConnection:
            self.connection_option = f"-R \"{relay_connection_name}:{host}:{';'.join(ports)}\""
        else:
            self.connection_option = f"-L \"{host}:{';'.join(ports)}:{relay_connection_name}\""

        self.az_relay_connection_wait_time = az_relay_connection_wait_time


    def __del__(self):
        """destructor
        """
        self.close()


    def az_relay_bridge_subprocess(self) -> subprocess.Popen:
        """Returns Azure Relay Bridge process subprocess.Popen object.
        None if one was not launched

        Returns:
            subprocess.Popen: Azure Relay Bridge process
        """
        return self.relay_subprocess


    def open(self, wait_for_connection: bool = True):
        """Launches Azure Relay Bridge tool with configured parameters
        (as initialized when creating DebugRelay object).
        If Azure Relay Bridge is not installed, installs it.

        Args:
            wait_for_connection (bool, optional): Wait for Azure Relay Bridge to initialize and connect. Defaults to True.

        Raises:
            TimeoutError: Raised when it takes longer than az_relay_connection_wait_time secods
                        for Azure Relay Bridge to initialize and connect.
        """
        # close existing Azure Relay Bridge process (if running)
        self.close()
        # install Azure Relay Bridge (if not yet)
        DebugRelay._install_azure_relay_bridge()

        command = f"{DebugRelay.relay_app_name} {self.connection_option} {self.auth_option}"
        if DebugRelay._relay_config_file is not None:
            command += f" -f \"{DebugRelay._relay_config_file}\""
        # start Azure Relay Bridge
        if not DebugRelay.is_windows:
            self.relay_subprocess = subprocess.Popen(
                command,
                preexec_fn=os.setpgrp,
                stdin=None, stderr=subprocess.STDOUT, stdout=subprocess.PIPE,
                shell=True, universal_newlines=True, close_fds=True)
        else:
            self.relay_subprocess = subprocess.Popen(
                command, 
                creationflag = subprocess.CREATE_NEW_PROCESS_GROUP,
                stdin=None, stderr=subprocess.STDOUT, stdout=subprocess.PIPE,
                shell=True, universal_newlines=True, close_fds=True)

        start = time.perf_counter()

        remote_forward_ready = False
        local_forward_ready = False
        over_timeout = False
        connected = False

        if self.relay_subprocess.poll() is not None:
            msg = "Azure Relay Bridge stopped too soon!"
            self.logger.critical(msg)
            raise RuntimeError(msg)

        # If recognizing Azure Relay Bridge connection status, parse its output.
        if wait_for_connection:
            # Iterate over Azure Relay Bridge output lines, 
            # looking for lines with "LocalForwardHostStart," and "RemoteForwardHostStart," to appear.
            for line in iter(self.relay_subprocess.stdout.readline, ''):
                self.logger.info(line)
                print(line)
                if self.relay_subprocess.poll() is not None:
                    self.logger.critical("Azure Relay Bridge stopped.")
                    break
                if wait_for_connection and not connected:
                    if line.find("LocalForwardHostStart,") != -1:
                        local_forward_ready = True
                    elif line.find("RemoteForwardHostStart,") != -1:
                        remote_forward_ready = True
                    if remote_forward_ready and local_forward_ready:
                        connected = True
                        msg = "Azure Relay Bridge is connected!"
                        self.logger.info(msg)
                        break

                    time_delta = time.perf_counter() - start
                    # did take too long to initialize and connect?
                    if time_delta > self.az_relay_connection_wait_time:
                        over_timeout = True
                        break
        else:
            msg = "Azure Relay Bridge is running!"
            self.logger.info(msg)

        # Handle over-timeout status
        if over_timeout:
            msg = f"Azure Relay Bridge took too long to connect."
            self.logger.critical(msg)
            self.close()
            raise TimeoutError(msg)
        elif self.relay_subprocess.poll() is None:
            threading.Thread(target=self._handle_output, daemon=True).start()
        else:
            msg = "Azure Relay Bridge stopped too soon!"
            self.logger.critical(msg)
            raise RuntimeError(msg)
        

    def close(self):
        """Stops Azure Relay Bridge process launched by this object
        """
        if self.relay_subprocess is not None:
            if self.is_running():
                self.logger.info("Closing Debug Relay...")
                if not DebugRelay.is_windows:
                    os.killpg(os.getpgid(self.relay_subprocess.pid), signal.SIGTERM)
                else:
                    os.kill(self.relay_subprocess.pid, signal.CTRL_C_EVENT)
                try:
                    self.relay_subprocess.wait(timeout=3)
                except subprocess.TimeoutExpired:
                    self.relay_subprocess.kill()
            self.relay_subprocess = None


    def background_launch(self) -> subprocess.Popen:
        """Launches Azure Relay Bridge process in detached mode
        Doesn't assign self.relay_subprocess, az_relay_bridge_subprocess() will return None.
        """
        # close existing Azure Relay Bridge process (if running)
        self.close()
        # install Azure Relay Bridge (if not yet)
        DebugRelay._install_azure_relay_bridge()

        command = f"{DebugRelay.relay_app_name} {self.connection_option} {self.auth_option}"
        if DebugRelay._relay_config_file is not None:
            command += f" -f \"{DebugRelay._relay_config_file}\""

        # start Azure Relay Bridge
        detached_relay_subprocess = subprocess.Popen(
            command,
            stdin=None, stderr=None, stdout=None,
            shell=True, close_fds=True)
        # wait a second
        time.sleep(1)
        if detached_relay_subprocess.poll() is not None:
            msg = f"Azure Relay Bridge failed to launch."
            self.logger.critical(msg)
            self.close()
        else:
            msg = "Azure Relay Bridge is running!"
            self.logger.info(msg)

        return detached_relay_subprocess


    def wait(self):
        if self.relay_subprocess is not None:
            self.relay_subprocess.wait()
            self.relay_subprocess = None


    def is_running(self) -> bool:
        if self.relay_subprocess is not None:
            if self.relay_subprocess.poll() is None:
                return True
            else:
                self.relay_subprocess = None
        return False


    def _handle_output(self):
        for line in iter(self.relay_subprocess.stdout.readline, ''):
            if line.find("Microsoft.Azure.Relay.Bridge.EventTraceActivity, exception = ") != -1:
                msg = f"[Azure Relay Bridge FAILURE]: {line}"
                self.logger.critical(msg)
                self.close()
                break
            else:
                self.logger.info(line)


    @staticmethod
    def from_config(config_file: str, 
                    debug_mode: DebugMode = DebugMode.WaitForConnection,
                    host:str = "127.0.0.1",
                    ports:typing.List[str] = ["5678"]) -> any:
        if os.path.exists(config_file):
            with open(config_file) as cfg_file:
                config = json.load(cfg_file)
                relay_connection_name = config["AZRELAY_CONNECTION_NAME"]
                conn_str = config["AZRELAY_CONNECTION_STRING"]
                return DebugRelay(
                    access_key_or_connection_string=conn_str,
                    relay_connection_name=relay_connection_name,
                    debug_mode=debug_mode,
                    host=host,
                    ports=ports)
        else:
            return None
    

    @staticmethod
    def from_environment(debug_mode: DebugMode = DebugMode.WaitForConnection,
                         host: str = "127.0.0.1",
                         ports: typing.List[str] = ["5678"]) -> any:
        relay_connection_name = os.environ.get("AZRELAY_CONNECTION_NAME")
        conn_str = os.environ.get("AZRELAY_CONNECTION_STRING")
        if not relay_connection_name or not conn_str:
            print("AZRELAY_CONNECTION_STRING and AZRELAY_CONNECTION_NAME variables must be assigned.")
            return None
        else:
            return DebugRelay(
                access_key_or_connection_string=conn_str,
                relay_connection_name=relay_connection_name,
                debug_mode=debug_mode,
                host=host,
                ports=ports)


    @staticmethod
    def kill_relays():
        """Kills all Azure Relay Bridge processes (azrelay) - no matter who and how launched them
        """
        if DebugRelay.is_windows:
            subprocess.run(
                f"taskkill /IM \"{DebugRelay.relay_app_name}.exe\" /F", shell=True)
        else:
            subprocess.run(f"pkill -9 {DebugRelay.relay_app_name}", shell=True)


    @staticmethod
    def _install_azure_relay_bridge():
        """Installs or updates Azure Relay Bridge
        """
        if DebugRelay._installed_az_relay:
            return
        DebugRelay._installed_az_relay = True

        azrelay_folder = os.path.join(
            Path.home(), DebugRelay.relay_dir_name, DebugRelay.relay_version_name)
        azrelay_parent = os.path.join(
            Path.home(), DebugRelay.relay_dir_name)
        azrelay_symlink = os.path.join(
            azrelay_parent, DebugRelay.relay_app_name)
        relay_file = os.path.join(
            azrelay_folder, DebugRelay.relay_app_name)
        DebugRelay._relay_config_file = os.path.join(
            azrelay_folder, DebugRelay.relay_app_name) + ".yml"

        exists = os.path.exists(azrelay_folder)
        if not exists:
            if DebugRelay.is_windows:
                download = DebugRelay.DEFAULT_AZ_RELAY_BRIDGE_WINDOWS_DOWLOAD
            else:
                plat = platform.platform().lower()
                if plat.startswith("macos"):
                    download = DebugRelay.DEFAULT_AZ_RELAY_BRIDGE_MACOS_DOWLOAD
                elif "-ubuntu" in plat or plat.startswith("ubuntu"):
                    download = DebugRelay.DEFAULT_AZ_RELAY_BRIDGE_UBUNTU_DOWLOAD
                else: # assume Debian
                    download = DebugRelay.DEFAULT_AZ_RELAY_BRIDGE_DEBIAN_DOWLOAD
                    if "debian" not in plat:
                        logging.warning(f"You are running an unsupported OS: {plat}. "\
                            "Using Debian build of Azure Relay Bridge.")
            
            ctx = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE

            filestream = urllib.request.urlopen(download, context=ctx)
            
            if download.lower().endswith(".zip"):
                with zipfile.ZipFile(filestream, 'r') as zip_ref:
                    zip_ref.extractall(azrelay_folder)
            else:
                with tarfile.open(fileobj=filestream, mode="r|gz") as thetarfile:
                    thetarfile.extractall(azrelay_folder)

            if not DebugRelay.is_windows:
                st = os.stat(relay_file)
                os.chmod(relay_file, st.st_mode | stat.S_IEXEC)

        if not exists or not os.path.exists(azrelay_symlink):
            tmp_link = f"{azrelay_symlink}.tmp"
            os.symlink(relay_file, tmp_link)
            os.replace(tmp_link, azrelay_symlink)
            st = os.stat(azrelay_symlink)
            os.chmod(azrelay_symlink, st.st_mode | stat.S_IEXEC)

        if not os.path.exists(DebugRelay._relay_config_file):
            with open(DebugRelay._relay_config_file, "a") as yml:
                yml.write("ExitOnForwardFailure: true")

        existing_path_var = os.environ["PATH"]
        paths = existing_path_var.split(os.pathsep)
        if azrelay_parent not in paths:
            os.environ["PATH"] += os.pathsep + azrelay_parent


def _main(connect: bool, host: str, ports: typing.List[str] = ["5678"], connection_string: str = None, relay_connection_name: str = None, config_file: str = None):
    """CLI main function

    Args:
        connect (bool): Connect (if True) or listen for incoming connections
        host (string): local hostname/address the debugging starts on (127.0.0.1)
        port (int): Azure Relay Bridge port
        connection_string (str): Optional connection string of an Azure Relay Hybrid Connection
        relay_connection_name (str): Optional hybrid connection name
        config_file (str): Optional configuration file path. Only used if connection_string is None.

    Raises:
        ValueError: Invalid arguments
        Exception: Cannot load configuration
    """
    print("Debug Relay Initialization...")

    mode = DebugMode.Connect if connect else DebugMode.WaitForConnection

    if connection_string is not None:
        if relay_connection_name is None:
            msg = "Both connection string and connection name must be provided."
            print(msg)
            raise ValueError(msg)
        debug_relay = DebugRelay(
            connection_string, relay_connection_name, mode, None, host, ports=ports)
    elif config_file is not None:
        if os.path.exists(config_file):
            debug_relay = DebugRelay.from_config(config_file, debug_mode=mode, host=host, ports=ports)
        else:
            config_file = os.path.normpath(config_file)
            logging.warning(f"Cannot load configuration file {config_file}. Trying with environment variables.")
            debug_relay = None
    else:
        debug_relay = None
    
    if debug_relay is None:
        debug_relay = DebugRelay.from_environment(
                debug_mode=mode, host=host, ports=ports)
    
    if debug_relay is None:
        raise Exception("Cannot create a Debug Relay object. Configuration may be missing.")

    print(f"Starting Debug relay...")
    relay = debug_relay.background_launch()
    relay.wait()


def _cli_main(argv):
    """CLI entry function

    Args:
        argv: Command Line arguments

        --no-kill - optional,
            If presented, prevents existing Azure Relay Bridge processes from being nuked.
            If omitted, all existing Azure Azure Relay Bridge processes will be killed.
        --mode - required,
            Debugging mode: listen, connect or none (default).
        --host - optional, defaults to 127.0.0.1, 
            Local hostname/address the debugging starts on (127.0.0.1)
        --ports - optional, defaults to 5678
            Azure Relay Bridge port
        --connection-string - optional, defaults to None
            Connection string of an Azure Relay Hybrid Connection
        --connection-name - optional, defaults to None
            Hybrid connection name. Required if --connection-string is specified.
        --config_file - optional, defaults to None
            Configuration file path. Only used if connection_string is not specified.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--no-kill', action='store_true',
                        default=False, required=False, help="Don't terminate existing azrelay processes.")
    parser.add_argument('--mode', action='store',
                        default="none", choices=['connect', 'listen', "none"], required=False,
                        help="Debugging mode: listen, connect or none")
    parser.add_argument('--ports', type=str,
                        default="5678", required=False, help="One or more Azure Relay Bridge ports.")
    parser.add_argument('--host', action='store',
                        default="127.0.0.1", required=False, help="Local hostname/address the debugging starts on")
    parser.add_argument('--connection-string', action='store',
                        default=None, required=False, help="Connection string of an Azure Relay Hybrid Connection")
    parser.add_argument('--connection-name', action='store',
                        default=None, required=False, help="Azure Relay Hybrid Connection name")
    parser.add_argument('--config-file', action='store',
                        default=None, required=False, help="Path to the configuration file. Defaults to None.")
    options = parser.parse_args(args=argv)

    logging.root.setLevel(logging.INFO)
    if not options.no_kill:
        print("Closing existing Azure Debug Relay processes.")
        DebugRelay.kill_relays()

    if options.mode != "none":
        connect = True if options.mode == "connect" else False
        ports = options.ports.strip()
        ports = ports.replace(", ", ",").replace(" ,", "").replace(" ", ",")
        ports_list = ports.split(",")
        _main(connect, options.host, ports_list, options.connection_string,
              options.connection_name, options.config_file)


# DebugRelays can work as a CLI tool.
if __name__ == '__main__':
    _cli_main(sys.argv[1:])
    
    
    



from .debug_relay import DebugRelay, DebugMode
from .debugpyex import DebugPyEx

__all__ = [
    "DebugRelay",
    "DebugMode",
    "debugpy_connect_with_timeout"
]


def debugpy_connect_with_timeout(host, port, connect_timeout_seconds):
    return DebugPyEx.connect(str(host), int(port), float(connect_timeout_seconds))


import ctypes
import inspect
import threading


def _async_raise(tid, exctype):
    '''Raises an exception in the threads with id tid'''
    if not inspect.isclass(exctype):
        raise TypeError("Only types can be raised (not instances)")
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid),
                                                     ctypes.py_object(exctype))
    if res == 0:
        raise ValueError("Invalid thread id")
    elif res != 1:
        # "if it returns a number greater than one, you're in trouble,
        # and you should call it again with exc=NULL to revert the effect"
        ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), None)
        raise SystemError("PyThreadState_SetAsyncExc failed")


class StoppableThread(threading.Thread):
    '''A thread class that supports raising an exception in the thread from
       another thread.
    '''
    def _get_my_tid(self):
        """determines this (self's) thread id

        CAREFUL: this function is executed in the context of the caller
        thread, to get the identity of the thread represented by this
        instance.
        """
        if not self.isAlive():
            raise threading.ThreadError("the thread is not active")

        # do we have it cached?
        if hasattr(self, "_thread_id"):
            return self._thread_id

        # no, look for it in the _active dict
        for tid, tobj in threading._active.items():
            if tobj is self:
                self._thread_id = tid
                return tid

        raise AssertionError("could not determine the thread's id")


    def raise_exception(self, exctype):
        """Raises the given exception type in the context of this thread.

        If the thread is busy in a system call (time.sleep(),
        socket.accept(), ...), the exception is simply ignored.

        If you are sure that your exception should terminate the thread,
        one way to ensure that it works is:

            t = ThreadWithExc( ... )
            ...
            t.raiseExc( SomeException )
            while t.isAlive():
                time.sleep( 0.1 )
                t.raiseExc( SomeException )

        If the exception is to be caught by the thread, you need a way to
        check that your thread has caught it.

        CAREFUL: this function is executed in the context of the
        caller thread, to raise an exception in the context of the
        thread represented by this instance.
        """
        _async_raise(self._get_my_tid(), exctype)


    def stop(self):
        self.raise_exception(SystemExit)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['azdebugrelay']

package_data = \
{'': ['*']}

install_requires = \
['debugpy>=1.2.1,<2.0.0']

setup_kwargs = {
    'name': 'azure-debug-relay',
    'version': '0.3.4',
    'description': 'Cross-network remote debugging for Python',
    'long_description': '# Azure Debug Relay for Python\n\nAzure Debug Relay - is a [Visual Studio Code](https://code.visualstudio.com/) extension and a Python package for distributed remote debugging.\n\n* [Azure Debug Relay extension](https://marketplace.visualstudio.com/items?itemName=VladKolesnikov-vladkol.azure-debug-relay) on Visual Studio Marketplace\n* [azure-debug-relay](https://pypi.org/project/azure-debug-relay/) package on PyPI\n\nAzure Debug Relay uses [debugpy](https://github.com/microsoft/debugpy) and [Azure Relay](https://docs.microsoft.com/en-us/azure/azure-relay/relay-what-is-it) service to create a debugging tunnel between 2 machines:\n\n1. You local Visual Studio Code debugger in `listen` mode.\n1. You remote code in `attach` mode.\n\nBoth machines can be isolated behind NAT or virtual networks - all they need is to be able to connect to Azure Relay resource.\nAzure Relay carries a secure tunnel, just as if these machines were in the same VPN.\n\n![Azure Relay Debugging Bridge](https://raw.githubusercontent.com/vladkol/azure-debug-relay/main/images/debug-relay-diagram.png)\n\nThe debugging tunnel is handled by **[Azure Relay Bridge](https://github.com/vladkol/azure-relay-bridge)** utility which is downloaded and installed automatically by Azure Debug Relay. Azure Relay Bridge can maintain secure TCP and UDP tunnels for different purposes.\n\n> We currently use a private fork of [Azure Relay Bridge](https://github.com/Azure/azure-relay-bridge) repo.\n\n## Requirements\n\n* Python 3.6+\n* debugpy 1.2.1+\n* Visual Studio Code 1.34+ (for using VS Code extension)\n\nAzure Relay Bridge tool is a .NET Core application, so you may need  to install `apt-transport-https` and other .NET Core 3.1 Runtime prerequisites on [Linux](https://docs.microsoft.com/en-us/dotnet/core/install/linux) and [Windows](https://docs.microsoft.com/en-us/dotnet/core/install/windows?tabs=netcore31).\n\n> You don\'t have to install .NET Runtime itself - Azure Relay Bridge builds are self-contained.\n\n### Supported Operating Systems\n\n* Ubuntu 18+\n* Debian 10+\n* macOS 10+\n* Windows 10\n\n## Installation\n\n**On the debugger side (usually your dev machine with Visual Studio code)**:\n\n> Install [Azure Debug Relay extension](https://marketplace.visualstudio.com/items?itemName=VladKolesnikov-vladkol.azure-debug-relay) from Visual Studio Marketplace.\n\n**On the server side**:\n\n> `python3 -m pip install azure-debug-relay`\n\n## Usage\n\nBefore you start debugging with Azure Debug Relay, there are 3 places you configure it:\n\n1. Azure Portal.\n1. Local machine where you run Visual Studio Code and its Python debugger.\n1. Remote machine where you run the same code files that open locally in VS Code.\n\n### In Azure Portal\n\n1. [Create Azure Relay resource](https://ms.portal.azure.com/#create/Microsoft.Relay). Better make one in a region closest to your location.\n1. Once created, switch to the resource, and select `Hybrid Connections` option in the vertical panel.\n1. Add a hybrid connection (`+ Hybrid Connection` button), give it a memorable name (e.g. `test` 🙂) - this is your **Hybrid Connection Name**.\n1. Switch to that new hybrid connection, then select `Shared Access Policies` in the vertical panel.\n1. Add a new policy with `Send` and `Listen` permissions.\n1. Once created, copy its `Primary Connection String`, this is your **Connection String**.\n\n#### **Azure CLI version**\n\nChoose your name instead of `mydebugrelay1` for an Azure Relay resource, and your custom name for Hybrid Connection instead of `debugrelayhc1`.\n\n```cmd\naz group create --name debugRelayResourceGroup --location westus2\n\naz relay namespace create --resource-group debugRelayResourceGroup --name mydebugrelay1 --location westus2\n\naz relay hyco create --resource-group debugRelayResourceGroup --namespace-name mydebugrelay1 --name debugrelayhc1\n\naz relay hyco authorization-rule create --resource-group debugRelayResourceGroup --namespace-name mydebugrelay1 --hybrid-connection-name debugrelayhc1 --name sendlisten --rights Send Listen\n\naz relay hyco authorization-rule keys list --resource-group debugRelayResourceGroup --namespace-name mydebugrelay1 --hybrid-connection-name debugrelayhc1 --name sendlisten\n```\n\nLast command will show you something like this:\n\n```json\n{\n  "keyName": "sendlisten",\n  "primaryConnectionString": "Endpoint=sb://mydebugrelay1.servicebus.windows.net/;SharedAccessKeyName=sendlisten;SharedAccessKey=REDACTED1;EntityPath=debugrelayhc1",\n  "primaryKey": "REDACTED1",\n  "secondaryConnectionString": "Endpoint=sb://mydebugrelay1.servicebus.windows.net/;SharedAccessKeyName=sendlisten;SharedAccessKey=REDACTED2;EntityPath=debugrelayhc1",\n  "secondaryKey": "REDACTED2"\n}\n```\n\nUse `primaryConnectionString` or `secondaryConnectionString` value as your **Connection String**.\n\n**Hybrid Connection Name** would be the one you choose instead of `debugrelayhc1`.\n</details>\n\n### Remotely with `remote_server_demo.py` or your code\n\nRemote Server example (in `samples/simple_demo/remote_server_demo.py`) assumes that Azure Relay credentials will are passes via `.azrelay.json` file in the current directory or via environment variables. Therefore, you have 2 options:\n\n**Option 1**: Create `.azrelay.json` file in your workspace directory root or whatever directory will be "current",\nand set 2 variables:\n\n1. `AZRELAY_CONNECTION_STRING` to your **Connection String**.\n1. `AZRELAY_CONNECTION_NAME` to your **Hybrid Connection Name**.\n\nFor example:\n\n```json\n{\n  "AZRELAY_CONNECTION_STRING": "Endpoint=sb://mydebugrelay1.servicebus.windows.net/;SharedAccessKeyName=sendlisten;SharedAccessKey=REDACTED1;EntityPath=debugrelayhc1",\n  "AZRELAY_CONNECTION_NAME": "debugrelayhc1"\n}\n```\n\nMake sure you add `.azrelay.json` to `.gitignore` so won\'t be committed.\n\n**Option 2**: You can assign these 2 variables as environment variables: `AZRELAY_CONNECTION_STRING` and `AZRELAY_CONNECTION_NAME` instead.\n\n### Prepare local Visual Studio Code\n\nUse `.azrelay.json` file in the root of your workspace as above or `.vscode/settings.json` with the following settings (actual values are ones you have):\n\n```json\n{\n  "azure-debug-relay.azrelay-connection-string": "Endpoint=sb://your-relay.servicebus.windows.net/;SharedAccessKeyName=key_name;SharedAccessKey=REDACTED;EntityPath=test",\n\n  "azure-debug-relay.azrelay-connection-name": "test",\n}\n```\n\n> Whenever Azure Debug Relay VS Code extension detects non-empty `azure-debug-relay.hybrid-connection-string` and `azure-debug-relay.hybrid-connection-name` settings (`vscode/settings.json`) or `AZRELAY_CONNECTION_STRING` and `AZRELAY_CONNECTION_NAME` in `.azrelay.json` file, it launches Azure Relay Bridge every time a debugging session with debugpy in `listen` mode is about to begin. If extension settings are not empty and `.azrelay.json` is present, Azure Relay Bridge prefers values from the extension settings (`vscode/settings.json`).\n\nVisual Studio Code extension ignores `AZRELAY_CONNECTION_STRING` and `AZRELAY_CONNECTION_NAME` environment variables.\n\n### Start debugging in Visual Studio Code\n\nThis step must be done on your dev machine in Visual Studio Code before launching the remote code.\n\n1. Open `remote_server_demo.py` and put a breakpoint in `do_work()` function.\n1. Makes sure your `.vscode/launch.json` has `Python: Listen 5678` configuration as in this repo\'s `.vscode/launch.json`.\n1. Start debugging in your local Visual Studio Code in `Python: Listen 5678` configuration.\n\nNotice how the debugger maps paths on the local and the remote machines.\nIf your code has a different structure remotely, you may need to provide more sophisticated path mappings. Here is that piece in `.vscode/launch.json`:\n\n```json\n"pathMappings": [\n    {\n        "localRoot": "${workspaceFolder}",\n        "remoteRoot": "."\n    }\n]\n```\n\nIt tells VS Code that the workspace directory locally is mapped to the "current" directory remotely.\n\nWhen the debugger looks goes through a file remotely, it needs to find the corresponding file in your local VS Code workspace.\nWhen debugging `remote_server_demo.py`, the debugger maps `./samples/simple_demo/remote_server_demo.py` remotely to `${workspaceFolder}/samples/simple_demo/remote_server_demo.py` locally.\n\n### Launch the example on the remote machine\n\n1. Clone the repo.\n1. Start `python3 ./samples/simple_demo/remote_server_demo.py --debug=attach`. Notice that current directory must contain `.azrelay.json` file unless configured with environment variables.\n\n> Terminal session where you start #2 must have the repo\'s directory as current directory - for a reason of mapping local and remote directories.\n\nIf everything works as it\'s supposed to, you will hit a breakpoint in your local Visual Studio Code.\n\n## Azure Debug Relay Python API\n\n`remote_server_demo.py` shows how you can use Azure Debug Relay (azure-debug-relay package) with your code.\n\n**azdebugrelay** module contains DebugRelay class that install and launches Azure Relay Bridge:\n\n```python\nfrom azdebugrelay import DebugRelay, DebugMode, debugpy_connect_with_timeout\n\naccess_key_or_connection_string = "AZURE RELAY HYBRID CONNECTION STRING OR ACCESS KEY"\nrelay_connection_name = "HYBRID CONNECTION NAME" # your Hybrid Connection name\ndebug_mode = DebugMode.Connect # or DebugMode.WaitForConnection if connecting from another end\nhybrid_connection_url = "HYBRID CONNECTION URL" # can be None if access_key_or_connection_string is a connection string\nhost = "127.0.0.1" # local hostname or ip address the debugger starts on\nport = 5678\nports = [str(port)] # any available ports that you can use within your machine\ndebugpy_timeout = 15 # 15 seconds for debugpy to connect\n\ndebug_relay = DebugRelay(access_key_or_connection_string, relay_connection_name, debug_mode, hybrid_connection_url, host, ports)\ndebug_relay.open()\n\n# attach to a remote debugger (usually from remote server code) with debug_mode = DebugMode.Connect\ndebugpy_connect_with_timeout(host, port, debugpy_timeout) # use instead of debugpy.connect\n# if debug_mode = DebugMode.WaitForConnection, we are going to listen instead\n# debugpy.listen((host, port))\n# if debug_mode = DebugMode.WaitForConnection, you can start DebugRelay on multiple ports (ports parameter is a list)\n# debugpy.listen must be called with each of these ports\n\n# Debug, debug, debug\n# ...\n# ...\n\ndebug_relay.close()\n```\n\n* `access_key_or_connection_string` - SAS Policy key or Connection String for Azure Relay Hybrid Connection. Must have `Send` and `Listen` permissions\n* `relay_connection_name` - name of the Hybrid Connection\n* `debug_mode` - debug connection mode. `DebugMode.WaitForConnection` when starting in listening mode, `DebugMode.Connect` for attaching to a remote debugger.\n* `hybrid_connection_url` - Hybrid Connection URL. Required when access_key_or_connection_string as an access key, otherwise is ignored and may be None.\n* `host` - Local hostname or ip address the debugger starts on, `127.0.0.1` by default\n* `port` - debugging port, `5678` by default\n\n> We added `debugpy_connect_with_timeout` method on top of **debugpy.connect()**.\nIt accepts `connect_timeout_seconds` parameter - how long it should wait for `debugpy.connect()` to connect.\nIf the connection is not successfully made within the timeout,\nthe debugging session aborts, and that can be handled in your code:\n`debugpy_connect_with_timeout()` returns `True` if the connection was successful, and `False` otherwise.\n\nNotice that DebugRelay accepts multiple ports to work with (**`ports` parameter is a list**).\nThat\'s because Azure Relay Bridge support forwarding on multiple ports.\nThis feature is primarily used by DebugRelay internally\nfor [Simultaneous distributed debugging](#simultaneous-distributed-debugging).\n\n### Azure Machine Learning samples\n\n**Simple Azure ML sample** is located in `samples/azure_ml_simple` directory.\n\nIt has 2 components:\n\n1. `deploy_and_run.py` script that deploys and launches an Azure ML pipeline with a single step.\n2. `steps/train.py` script which contains that simple step.\n\nLook at the [sample\'s readme file](samples/azure_ml_simple/README.md).\n\n**Advanced Azure Machine Learning sample** is located in `samples/azure_ml_advanced`, and demonstrates a complex debugging scenario with parallel steps.\n\nLook at the [advanced sample\'s readme file](samples/azure_ml_advanced/README.md).\n\n## Simultaneous distributed debugging\n\nYou can debug multiple simultaneously running remote nodes using different ports.\nEach execution flow you want to debug must use a separate port - both locally and remotely.\n\nIn Visual Studio code, it is achievable via so called "[compound launch configurations](https://code.visualstudio.com/docs/editor/debugging#_compound-launch-configurations)".\nCompound launch configurations combine multiple launch configurations, and therefore start them at the same time.\n\nEach launch configuration must be a Python `listen` configuration with a unique name and port:\n\n```json\n{\n  "name": "Python: Listen 5678",\n  "type": "python",\n  "request": "attach",\n  "listen": {\n      "host": "127.0.0.1",\n      "port": 5678\n  },\n  "pathMappings": [\n      {\n          "localRoot": "${workspaceFolder}",\n          "remoteRoot": "."\n      }\n  ]\n}\n```\n\nYou need as many launch configurations as number of simultaneous execution flows or nodes you\'d like to debug.\nThen you combine them in `.vscode/launch.json` to as a compound:\n\n```json\n"compounds": [\n  {\n    "name": "Python: AML Adv 2 Listeners",\n    "configurations": [\n      "Python: Listen 5678", \n      "Python: Listen 5679"\n    ]\n  }\n]\n```\n\nRemotely, each node you debug should be aware of the port number it should use.\nThat port number must be passed to `DebugRelay` object and `debugpy_connect_with_timeout()`.\n\n## Troubleshooting\n\nWhy using [Azure Relay Bridge](https://github.com/Azure/azure-relay-bridge) which is a .NET Core application that we have to install and use via `subprocess` calls?\n\nReasons:\n\n1. Azure Relay has SDKs for .NET, Java, and Node. [No Python SDK or examples](https://github.com/Azure/azure-relay/issues/28#issuecomment-390778193).\n1. Azure Relay Bridge does a lot of things we have to implement otherwise. It is a great tool that can help you connecting different networks for many purposes: for RDP, SSH and other protocols over TCP or UDP.\n\nA [private fork](https://github.com/vladkol/azure-relay-bridge) we are currently using is only to provide .NET Core 3.1 builds of the most recent code. There is a pending pul-requests: [one](https://github.com/Azure/azure-relay-bridge/pull/22) and [two](https://github.com/Azure/azure-relay-bridge/pull/19).\n\n### Known issues\n\n> **Extension and python module fail complaining about `AZRELAY_CONNECTION_NAME` or `azrelay-connection-name`**.\n\n**Reason**: It is a breaking change in 0.2.0. We changed configuration and environment variables name from `AZRELAY_NAME` to `AZRELAY_CONNECTION_NAME`. VS Code extension setting names we also renamed:\n\n* `hybrid-connection-string` to **azrelay-connection-string**\n* `azrelay-connection-name` to **azrelay-connection-name**\n\n**Workaround**: Modify your `.azrelay.json`, environment variables or `.vscode/settings.json` files accordingly.\n\n> **On macOS, there may be a situation when Azure Relay Bridge (`azbridge`) cannot connect when creating a local forwarder** (`-L` option).\n\n**Reason**: .NET Core wants you to add your Computer Name to `/etc/hosts` file, and make sure the hostname is configured.\n\n**Workaround**: Make necessary edits of `/etc/hosts` file, and configure hostname:\n\n1. Look for your computer\'s name in `Settings → Sharing`.\n1. Run the following command: `scutil --set HostName "your-computer-name"`. (**replace `your-computer-name` with your computer\'s name**)\n1. Open `/etc/hosts` in a text editor in *sudo* mode (VS Code can save it later in *sudo* mode).\n1. Add the following line (**replace `your-computer-name` with your computer\'s name**). Save the file.\n\n```text\n127.0.0.1   your-computer-name\n```\n\n> **I launched the debugger as described and nothing happened**\n\n**Reason**: you *probably* didn\'t put a breakpoint in your VS Code locally. Make sure that breakpoint is in a place that your server process actually runs through.\n\n> **I do everything right, but nothing works**\n\n**Reason**: Stop all debugging sessions (if any). Kill all `azbridge` processes locally and remotely. Try again.\n\nDoesn\'t help? [File an issue](https://github.com/vladkol/azure-debug-relay/issues/new)! Thank you!\n',
    'author': 'Vlad Kolesnikov',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
